local core = core or {}
local data = {}
local func_handles = {}

local unpack = unpack or table.unpack

local function packn(...)
    return { n = select('#', ...), ... }
end

local function unpackn(t)
    return unpack(t, 1, t.n)
end

local function mergen(...)
    local res = { n = 0 }
    for i = 1, select('#', ...) do
        local t = select(i, ...)
        for j = 1, t.n do
            res.n = res.n + 1
            res[res.n] = t[j]
        end
    end
    return res
end

local function trackHandle(handle)
    if handle.obj._id ~= nil then
        if func_handles[handle.obj._id] == nil then
            func_handles[handle.obj._id] = setmetatable({}, { __mode = 'v' })
        end
        table.insert(func_handles[handle.obj._id], handle)
    end
end

-- Table functions curtosy of http://lua-users.org/wiki/TableUtils
function table.val_to_str(v)
    if "string" == type(v) then
        v = string.gsub(v, "\n", "\\n")
        if string.match(string.gsub(v, "[^'\"]", ""), '^"+$') then
            return "'" .. v .. "'"
        end
        return '"' .. string.gsub(v, '"', '\\"') .. '"'
    else
        return "table" == type(v) and table.tostring(v) or
            tostring(v)
    end
end

function table.key_to_str(k)
    if "string" == type(k) and string.match(k, "^[_%a][_%a%d]*$") then
        return k
    else
        return "[" .. table.val_to_str(k) .. "]"
    end
end

function table.tostring(tbl)
    local result, done = {}, {}
    for k, v in ipairs(tbl) do
        table.insert(result, table.val_to_str(v))
        done[k] = true
    end
    for k, v in pairs(tbl) do
        if not done[k] then
            table.insert(result,
                table.key_to_str(k) .. "=" .. table.val_to_str(v))
        end
    end
    return "{" .. table.concat(result, ",") .. "}"
end

core.cb = function(obj, funcName)
    local handle = {}
    assert(obj[funcName], table.tostring(obj) .. " does not contain function '" .. funcName .. "'");
    if obj.prototype then
        handle.callback = function(self, ...) return obj[funcName](...) end
        trackHandle(handle)
    else
        handle.callback = function(self, ...) return obj[funcName](obj, ...) end
    end
    setmetatable(handle,
        { __call = handle.callback, __tostring = function(self) return "core.cb(" .. funcName .. ")" end })
    return handle
end

core.bind = function(obj, funcName, ...)
    -- TODO(jmcmorris): Should values in handle be weak such as obj?
    local handle = { }
    local args = packn(...)
    if obj.prototype then
        handle.callback = function(self, ...) return obj[funcName](unpackn(mergen(args, packn(...)))) end
        trackHandle(handle)
    else
        handle.callback = function(self, ...) return obj[funcName](obj, unpackn(mergen(args, packn(...)))) end
    end
    setmetatable(handle, { __call = handle.callback })
    return handle
end

core.instantiate = function(class)
    print(class)
    print(class._id)
    for key, value in pairs(class) do
        print('\t', key, value)
    end
    print("core.instantiate(" .. class._id .. ")")
    local obj = setmetatable({ _isinstance = true }, class.prototype)
    if data[class._id] == nil then
        data[class._id] = setmetatable({}, { __mode = 'v' })
    end
    table.insert(data[class._id], obj)
    return obj
end

core.reload = function(class, reloadFunc)
    if data[class._id] then
        local instances = data[class._id]
        for i = 1, #instances do
            if instances[i] == nil then
                table.remove(instances, i)
            else
                local instance = instances[i]
                setmetatable(instance, { __index = class })
                if reloadFunc ~= nil then
                    reloadFunc(instance)
                end
            end
        end
    end
    if func_handles[class._id] then
        local handles = func_handles[class._id]
        for i = 1, #handles do
            if handles[i] == nil then
                table.remove(handles, i)
            else
                handles[i].obj = class
            end
        end
    end
end

if not HOT_RELOADING then
    core.cb = function(obj, funcName)
        if obj.prototype == nil then
            return function(...) return obj[funcName](obj, ...) end
        end
        return obj[funcName]
    end

    core.bind = function(obj, funcName, ...)
        -- TODO(jmcmorris): Should values in handle be weak such as obj?
        local args = packn(...)
        local handle = { }
        if obj.prototype then
            handle.callback = function(self, ...) return obj[funcName](unpackn(mergen(args, packn(...)))) end
        else
            handle.callback = function(self, ...) return obj[funcName](obj, unpackn(mergen(args, packn(...)))) end
        end
        setmetatable(handle, { __call = handle.callback })
        return handle
    end

    core.instantiate = function(class)
        return setmetatable({ _isinstance = true }, { __index = class })
    end

    core.reload = function(class, reloadFunc)
    end
end

return { core = core }
